#ifndef DUNE_REDUCEDROD_CELLSOLVER_HH
#define DUNE_REDUCEDROD_CELLSOLVER_HH

#include <dune/reducedrod/cellassembler.hh>
#include <dune/functions/functionspacebases/hierarchicvectorwrapper.hh>

#include <dune/istl/operators.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/spqr.hh>



template <class Basis>
class CellSolver {

public:

	static const int nCompo = 3;

	using VectorRT = typename CellAssembler<Basis>::VectorRT;

	using VectorCT = typename CellAssembler<Basis>::VectorCT;
	using MatrixCT = typename CellAssembler<Basis>::MatrixCT;

	using HierarchicVectorView = Dune::Functions::HierarchicVectorWrapper<VectorCT, double>;

protected:

	CellAssembler<Basis>& cellAssembler_;

	MatrixCT stiffnessMatrix_;
	VectorCT load_a_, load_K1_, load_K2_, load_K3_;

public:

	// Constructor
	CellSolver(CellAssembler<Basis>& cellAssembler)
        : cellAssembler_(cellAssembler)
    {
    	cellAssembler.assemble(stiffnessMatrix_,load_a_, load_K1_, load_K2_, load_K3_); //assemble hier rein?
    }

    //return ref ???
    CellAssembler<Basis> getCellAssembler(){return cellAssembler_;}

    MatrixCT * getStiffnessMatrix(){return &stiffnessMatrix_;}
    VectorCT * getLoad_a(){return &load_a_;}
    VectorCT * getLoad_K1(){return &load_K1_;}
    VectorCT * getLoad_K2(){return &load_K2_;}
    VectorCT * getLoad_K3(){return &load_K3_;}

	void solveCorrectorSystems(VectorCT& x_a, VectorCT& x_K1, VectorCT& x_K2, VectorCT& x_K3)
	{

		std::cout << "CellSolver: Solve linear systems begin.\n";
		ParameterTree parameterSet = cellAssembler_.getParameterSet();
		//auto logPointer = cellAssembler_.getLog();
		//auto & log = *logPointer;
		auto & log = *(cellAssembler_.getLog());

		int sizeBasisTorus_ = cellAssembler_.getSizeTorusElements();
		x_a.resize(sizeBasisTorus_);
		x_K1.resize(sizeBasisTorus_);
		x_K2.resize(sizeBasisTorus_);
		x_K3.resize(sizeBasisTorus_);
		x_a = 0;
		x_K1 = 0;
		x_K2 = 0;
		x_K3 = 0;

		if (parameterSet.get<int>("iterate_with_load") == 1){
			x_a = load_a_;
			x_K1 = load_K1_;
			x_K2 = load_K2_;
			x_K3 = load_K3_;
		}

		if (parameterSet.get<int>("useCG") == 1){
			std::cout << "We use CG solver.\n";

		    MatrixAdapter<MatrixCT, VectorCT, VectorCT> op(stiffnessMatrix_);

		    std::cout << "MatrixAdapter op.\n";

		    // Sequential incomplete LU decomposition as the preconditioner
		    SeqILU<MatrixCT, VectorCT, VectorCT> ilu0(stiffnessMatrix_,1.0);
		    int iter = parameterSet.get<double>("iterations_CG", 999);
		    // Preconditioned conjugate-gradient solver
		    CGSolver<VectorCT> solver(op,
		                            ilu0, //NULL,
		                            1e-8, // desired residual reduction factor
		                          iter,   // maximum number of iterations
		                            2);   // verbosity of the solver
		    InverseOperatorResult statistics;
		    std::cout << "solve linear system for xa.\n";
		    solver.apply(x_a, load_a_, statistics);
		    std::cout << "solve linear system for xK1.\n";
		    solver.apply(x_K1, load_K1_, statistics);
		    std::cout << "solve linear system for xK2.\n";
		    solver.apply(x_K2, load_K2_, statistics);
		    std::cout << "solve linear system for xK3.\n";
		    solver.apply(x_K3, load_K3_, statistics);

		}

		if (parameterSet.get<int>("useQR") == 1){

		    std::cout << "We use QR solver.\n";
		    log << "solveLinearSystems: We use QR solver.\n";
		    //TODO install suitesparse
		    SPQR<MatrixCT> sPQR(stiffnessMatrix_);
		    sPQR.setVerbosity(1);
		    InverseOperatorResult statisticsQR;

		    sPQR.apply(x_a, load_a_, statisticsQR);
		    sPQR.apply(x_K1, load_K1_, statisticsQR);
		    sPQR.apply(x_K2, load_K2_, statisticsQR);
		    sPQR.apply(x_K3, load_K3_, statisticsQR);

		}

		// Write solutions in logs
		if (parameterSet.get<int>("write_solutions_corrector_problems") == 1){
		    log << "\nSolution of Corrector problems:\n";

		    auto sizeComp = x_a.size()/nCompo;
		    std::vector<VectorRT> x_a_vec(sizeComp);
		    std::vector<VectorRT> x_K1_vec(sizeComp);
		    std::vector<VectorRT> x_K2_vec(sizeComp);
		    std::vector<VectorRT> x_K3_vec(sizeComp);

	    	for (int i=0; i < x_a_vec.size(); i++){
	    		x_a_vec[i] = VectorRT{x_a[i], x_a[i + sizeComp], x_a[i + 2*sizeComp]};
	    		x_K1_vec[i] = VectorRT{x_K1[i], x_K1[i + sizeComp], x_K1[i + 2*sizeComp]};
	    		x_K2_vec[i] = VectorRT{x_K2[i], x_K2[i + sizeComp], x_K2[i + 2*sizeComp]};
	    		x_K3_vec[i] = VectorRT{x_K3[i], x_K3[i + sizeComp], x_K3[i + 2*sizeComp]};
	    	}

	    	log << "\nxa:\n";
		    for (int i=0; i < x_a_vec.size(); i++)
		      log << i << ": " << x_a_vec[i] << std::endl;

		    log << "\nxK1:\n";
		    for (int i=0; i < x_K1_vec.size(); i++)
		      log << i << ": " << x_K1_vec[i] << std::endl;

		    log << "\nxK2:\n";
		    for (int i=0; i < x_K2_vec.size(); i++)
		      log << i << ": " << x_K2_vec[i] << std::endl;

		    log << "\nxK3:\n";
		    for (int i=0; i < x_K3_vec.size(); i++)
		      log << i << ": " << x_K3_vec[i] << std::endl;

	    	/*
		    log << "\nxa:\n";
		    for (int i=0; i < x_a.size(); i++)
		      log << i << " " << x_a[i] << std::endl;

		    log << "\nxK1:\n";
		    for (int i=0; i < x_K1.size(); i++)
		      log << i << " " << x_K1[i] << std::endl;

		    log << "\nxK2:\n";
		    for (int i=0; i < x_K2.size(); i++)
		      log << i << " " << x_K2[i] << std::endl;

		    log << "\nxK3:\n";
		    for (int i=0; i < x_K3.size(); i++)
		      log << i << " " << x_K3[i] << std::endl;
		    */
		}

	}


	void writeSystemMatrixAndRHSs()
	{
  	  ParameterTree parameterSet = cellAssembler_.getParameterSet();
  	  auto logP = cellAssembler_.getLog();
	  auto & log = *logP;

	  if (parameterSet.get<int>("write_stiffness_matrix_Cell") == 1){
	    std::cout << "write stiffness matrix of cell problem in logfile\n";

	    log << "\nwriteSystemMatrixAndRHSs:\n";
	    printSparseMatrix(log, stiffnessMatrix_, "stiffnessMatrix with periodic bc", "");
	    //log << "\n";
	    //printmatrix(log, stiffnessMatrix_, "stiffnessMatrix with periodic bc", "");
	    //writeMatrixToMatlab(stiffnessMatrix, "../outputs/reducedrod_output/postprocessing/stiffness_matrix_for_octave.txt");
	  }

	  if (parameterSet.get<int>("write_loads_Cell") == 1){

	    std::cout << "write rhs of cell problem in logfile\n";

	    auto sizeComp = load_a_.size()/nCompo;
		std::vector<VectorRT> load_a_vec(sizeComp);
		std::vector<VectorRT> load_K1_vec(sizeComp);
		std::vector<VectorRT> load_K2_vec(sizeComp);
		std::vector<VectorRT> load_K3_vec(sizeComp);

	    for (int i=0; i < load_a_vec.size(); i++){
	    	load_a_vec[i] = VectorRT{load_a_[i], load_a_[i + sizeComp], load_a_[i + 2*sizeComp]};
	    	load_K1_vec[i] = VectorRT{load_K1_[i], load_K1_[i + sizeComp], load_K1_[i + 2*sizeComp]};
	    	load_K2_vec[i] = VectorRT{load_K2_[i], load_K2_[i + sizeComp], load_K2_[i + 2*sizeComp]};
	    	load_K3_vec[i] = VectorRT{load_K3_[i], load_K3_[i + sizeComp], load_K3_[i + 2*sizeComp]};
	    }

	    log << "\n RHS stretch strain - load_a:\n";
	    for (int i=0; i < load_a_vec.size(); i++)
	      log << i << " " << load_a_vec[i] << std::endl;

	    log << "\n RHS twist strain - load_K1:\n";
	    for (int i=0; i < load_K1_vec.size(); i++)
	      log << i << " " << load_K1_vec[i] << std::endl;

	    log << "\n RHS bend1 strain - load_K2:\n";
	    for (int i=0; i < load_K2_vec.size(); i++)
	      log << i << " " << load_K2_vec[i] << std::endl;

	    log << "\n RHS bend2 strain - load_K3:\n";
	    for (int i=0; i < load_K3_vec.size(); i++)
	      log << i << " " << load_K3_vec[i] << std::endl;

	  }
	}


	void plotLoads()
	{
		auto basis = cellAssembler_.getBasis();
		auto perIndexMap = cellAssembler_.getIndexMapPeriodicBoundary();
	    auto size = basis->size();

	    VectorCT load_a_Ce(size);
	    VectorCT load_K1_Ce(size);
	    VectorCT load_K2_Ce(size);
	    VectorCT load_K3_Ce(size);
	    //index_PeriodicFE = indexmap(index_FE)
	    for (int i = 0; i < size; i++){
	  		auto idx = perIndexMap[i];
	    	load_a_Ce[i] = load_a_[idx];
	    	load_K1_Ce[i] = load_K1_[idx];
	    	load_K2_Ce[i] = load_K2_[idx];
	    	load_K3_Ce[i] = load_K3_[idx];
	    }

	    std::cout << "plot loads.\n";

	    auto stretchF = Functions::makeDiscreteGlobalBasisFunction<VectorRT>(Functions::subspaceBasis(*basis), HierarchicVectorView(load_a_Ce));
	    auto twistF = Functions::makeDiscreteGlobalBasisFunction<VectorRT>(Functions::subspaceBasis(*basis), HierarchicVectorView(load_K1_Ce));
	    auto bend1F = Functions::makeDiscreteGlobalBasisFunction<VectorRT>(Functions::subspaceBasis(*basis), HierarchicVectorView(load_K2_Ce));
	    auto bend2F = Functions::makeDiscreteGlobalBasisFunction<VectorRT>(Functions::subspaceBasis(*basis), HierarchicVectorView(load_K3_Ce));

	    //  Write result to VTK file, Maybe we write for every gamma?
	    VTKWriter<typename Basis::GridView> vtkWriter(basis->gridView());
	    vtkWriter.addVertexData(stretchF, VTK::FieldInfo("load_a_stretch", VTK::FieldInfo::Type::vector, nCompo));
	    vtkWriter.addVertexData(twistF, VTK::FieldInfo("load_K1_twist", VTK::FieldInfo::Type::vector, nCompo));
	    vtkWriter.addVertexData(bend1F, VTK::FieldInfo("load_K2_bend1", VTK::FieldInfo::Type::vector, nCompo));
	    vtkWriter.addVertexData(bend2F, VTK::FieldInfo("load_K3_bend2", VTK::FieldInfo::Type::vector, nCompo));

	    std::string outputPath = cellAssembler_.getParameterSet().get("outputPath", "../outputs/clampedprestressedrod_output");
	    vtkWriter.write(outputPath + "/loads_cell_problems");
	}

}; // end class



#endif
