#ifndef DUNE_REDUCEDROD_EFFECTIVEMODELCOMPUTER_HH
#define DUNE_REDUCEDROD_EFFECTIVEMODELCOMPUTER_HH


#include <dune/reducedrod/cellsolver.hh>


template <class Basis>
class EffectiveModelComputer {

public:

	static const int nCompo = 3;
	static const int nCells = 4;

	using Domain = typename CellAssembler<Basis>::Domain;

	using VectorRT = typename CellSolver<Basis>::VectorRT;
	using MatrixRT = typename CellAssembler<Basis>::MatrixRT;

	using Func2Tensor = typename CellAssembler<Basis>::Func2Tensor;

	using VectorCT = typename CellSolver<Basis>::VectorCT;

	using HierarchicVectorView = typename CellSolver<Basis>::HierarchicVectorView;

protected:

	CellSolver<Basis>& cellSolver_;
	Func2Tensor prestrain_;

	VectorCT x_a_, x_K1_, x_K2_, x_K3_; // in torus basis
	FieldMatrix<double, nCells, nCells> M_Ce_;
	FieldVector<double, nCells> aeffKeff_;

public:

	// Constructor
	EffectiveModelComputer(CellSolver<Basis>& cellSolver, Func2Tensor prestrain)
        : cellSolver_(cellSolver), prestrain_(prestrain)
    {
    	cellSolver.solveCorrectorSystems(x_a_, x_K1_, x_K2_, x_K3_ );
    	computeEffectiveModel();
    	//computeCorrector();
    }

	VectorCT * getCorr_a(){return &x_a_;}
    VectorCT * getCorr_K1(){return &x_K1_;}
    VectorCT * getCorr_K2(){return &x_K2_;}
    VectorCT * getCorr_K3(){return &x_K3_;}

	FieldMatrix<double, nCells, nCells> getEffMaterial(){return M_Ce_;}
	FieldVector<double, nCells> getEffPreStrain(){return aeffKeff_;}

	CellSolver<Basis> getCellSolver(){return cellSolver_;}
	const Basis* getCellBasis()  {
		auto cellAssembler = cellSolver_.getCellAssembler();
		return cellAssembler.getBasis();}

	void plotCorrectors()
	{
		std::cout << "Write corrector solutions.\n";

		auto cellAssembler = cellSolver_.getCellAssembler();
		auto basis = getCellBasis();
		auto perIndexMap = cellAssembler.getIndexMapPeriodicBoundary();
	    auto size = basis->size();

	    VectorCT x_a_Ce(size);
	    VectorCT x_K1_Ce(size);
	    VectorCT x_K2_Ce(size);
	    VectorCT x_K3_Ce(size);
	    //index_PeriodicFE = indexmap(index_FE)
	    for (int i = 0; i < size; i++){
	  		auto idx = perIndexMap[i];
	    	x_a_Ce[i] = x_a_[idx];
	    	x_K1_Ce[i] = x_K1_[idx];
	    	x_K2_Ce[i] = x_K2_[idx];
	    	x_K3_Ce[i] = x_K3_[idx];
	    }

	    VectorCT x_corrector(size);
	    computeCorrector(x_corrector);

	    auto stretchF = Functions::makeDiscreteGlobalBasisFunction<VectorRT>(Functions::subspaceBasis(*basis), HierarchicVectorView(x_a_Ce));
	    auto twistF = Functions::makeDiscreteGlobalBasisFunction<VectorRT>(Functions::subspaceBasis(*basis), HierarchicVectorView(x_K1_Ce));
	    auto bend1F = Functions::makeDiscreteGlobalBasisFunction<VectorRT>(Functions::subspaceBasis(*basis), HierarchicVectorView(x_K2_Ce));
	    auto bend2F = Functions::makeDiscreteGlobalBasisFunction<VectorRT>(Functions::subspaceBasis(*basis), HierarchicVectorView(x_K3_Ce));
	  	auto correctorF = Functions::makeDiscreteGlobalBasisFunction<VectorRT>(Functions::subspaceBasis(*basis), HierarchicVectorView(x_corrector));

	    //  Write result to VTK file, Maybe we write for every gamma?
	    VTKWriter<typename Basis::GridView> vtkWriter(basis->gridView());
	    vtkWriter.addVertexData(stretchF, VTK::FieldInfo("phi_a_stretch", VTK::FieldInfo::Type::vector, nCompo));
	    vtkWriter.addVertexData(twistF, VTK::FieldInfo("phi_K1_twist", VTK::FieldInfo::Type::vector, nCompo));
	    vtkWriter.addVertexData(bend1F, VTK::FieldInfo("phi_K2_bend1", VTK::FieldInfo::Type::vector, nCompo));
	    vtkWriter.addVertexData(bend2F, VTK::FieldInfo("phi_K3_bend2", VTK::FieldInfo::Type::vector, nCompo));
	    vtkWriter.addVertexData(correctorF, VTK::FieldInfo("phi_full", VTK::FieldInfo::Type::vector, nCompo));

	    std::string outputPath = cellAssembler.getParameterSet().get("outputPath", "../outputs/clampedprestressedrod_output");
	    vtkWriter.write(outputPath + "/corrector_solutions");

	    //add corrector
	}


	void computeCorrector(VectorCT& x_corrector)
	{
		VectorCT x_Corrector_Torus(x_a_);
	    x_Corrector_Torus *= aeffKeff_[0];
	    auto tmp_xK1 = x_K1_;
	    auto tmp_xK2 = x_K2_;
	    auto tmp_xK3 = x_K3_;
	    tmp_xK1 *= aeffKeff_[1];
	    tmp_xK2 *= aeffKeff_[2];
	    tmp_xK3 *= aeffKeff_[3];
	    x_Corrector_Torus += tmp_xK1;
	    x_Corrector_Torus += tmp_xK2;
	    x_Corrector_Torus += tmp_xK3;

	    auto cellAssembler = cellSolver_.getCellAssembler();
		auto basis = cellAssembler.getBasis();
		auto perIndexMap = cellAssembler.getIndexMapPeriodicBoundary();

	    x_corrector.resize(basis->size());
	    for (int i = 0; i < basis->size(); i++)
	      x_corrector[i] = x_Corrector_Torus[perIndexMap[i]];
	}


private:
	void computeEffectiveModel()
	{
		auto cellAssembler = cellSolver_.getCellAssembler();
		ParameterTree parameterSet = cellAssembler.getParameterSet();
		auto logP = cellAssembler.getLog();
	  	auto & log = *logP;

		// lateral stretch strain	E_U1 = e1 ot e1
	  	Func2Tensor E_a = [] (const Domain& z) //extra Klasse
	  	{
			return MatrixRT{{1.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
	  	};

	  	// twist in x2-x3 plane E_K1 = sym (e1 x (0,x2,x3)^T ot e1)
	    Func2Tensor E_K1 = [] (const Domain& z)
	    {
	      	return MatrixRT{{0.0,-0.5*z[2], 0.5*z[1]}, {-0.5*z[2], 0.0 , 0.0 }, {0.5*z[1],0.0,0.0}};
	  	};

	    //  bend strain in x1-x2 plane	E_K2 = sym (e2 x (0,x2,x3)^T ot e1)
	  	Func2Tensor E_K2 = [] (const Domain& z)
	  	{
	      	return MatrixRT{{z[2], 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0}};
	  	};

	  	// bend strain in x1-x3 plane	E_K3 = sym (e3 x (0,x2,x3)^T ot e1)
	  	Func2Tensor E_K3 = [] (const Domain& z)
	  	{
	      	return MatrixRT{{-z[1], 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
	  	};

		VectorCT *xCorrectors[nCells] = {&x_a_, &x_K1_, &x_K2_, &x_K3_};
	  	VectorCT *loadsCell[nCells] = {cellSolver_.getLoad_a(), cellSolver_.getLoad_K1(), cellSolver_.getLoad_K2(), cellSolver_.getLoad_K3()};

	  	//using GridViewFunction_Ce = decltype(E_a);
	  	const std::array<Func2Tensor*, nCells> strainsE = {&E_a, &E_K1, &E_K2, &E_K3};

	  	//wir brauchen sym grad phi als gridFunction!

	  	FieldMatrix<double, nCells, nCells> QEE(0);

	  	// Assemble 4x4 matrix                                                                                             //symmetrisch? reicht hälfte zu füllen?
	  	for (unsigned int k = 0; k < nCells; k++)
	  		for (unsigned int l = 0; l < nCells; l++){
	  	  		auto energyEkEl = cellAssembler.energy(*strainsE[k], *strainsE[l]); //<LE,E>
	  	  		QEE[k][l] = energyEkEl;
	  	  		M_Ce_[k][l]= 3.0*((*xCorrectors[k])*(*loadsCell[l])) + energyEkEl; //<LChi, Chi>= <E,Chi>, 2<LChi, E>                         // 3* ?
	  		}// oder braucht man nicht periodische objekte?

	  	/*
	  	FieldMatrix<double, nCells, nCells> QChiChi(0);
	  	FieldMatrix<double, nCells, nCells> QEChi(0);
	  	if (parameterSet.get<bool>("material_computation_2"))
	  		for (unsigned int k = 0; k < nCells; k++)
	  			for (unsigned int l = 0; l < nCells; l++){
	  	  			auto energyEE = cellAssembler.computeScalarTerm(*strainsE[k], *strainsE[l]); //<LE,E>
	  	  			auto energyChiChi = cellAssembler.computeScalarTerm(*xCorrectors[k], *xCorrectors[l]);
	  	  			auto energyEChi = cellAssembler.computeScalarTerm(*strainsE[k], *xCorrectors[l]);
	  	  			QEE[k][l] = energyEE;
	  	  			QChiChi[k][l] = energyChiChi;
	  	  			QEChi[k][l] = energyEChi;
	  	  			M_Ce_[k][l]= energyEE + energyChiChi + 2*energyEChi;  //<LChi, Chi>= <E,Chi>, 2<LChi, E>                 //  eher so ??
	  			}
		*/



	  	VectorCT load_B;
	  	cellAssembler.assembleLoadVector(prestrain_, load_B);

	  	// Assemble rhs by projecting prestrain to strain basis
	    FieldVector<double, nCells> b;
	  	for (unsigned int k = 0; k < nCells; k++){
	  		auto strainsEkB = cellAssembler.energy(*strainsE[k], prestrain_); // <LB, E>
	  		b[k]= (*xCorrectors[k]) * load_B + strainsEkB; // < B, chi>
	  	}

	  	M_Ce_.solve( aeffKeff_ ,b);

	  	log << "\nQEE= \n";
	  	for (int i = 0; i < nCells; i++)
	  	  log << QEE[i] << std::endl;
	  	log << "\nM_Ce= \n";
	  	for (int i = 0; i < nCells; i++)
	  	  log << M_Ce_[i] << std::endl;
	  	log << "\nb= " << b << std::endl;
	  	log << "\nstretch twist bend1 bend2\naeffKeff = "<< aeffKeff_ << std::endl;
	  	std::cout << "\naeffKeff: stretch twist bend1 bend2 = "<< aeffKeff_ << std::endl;

	}

}; // end class





#endif
